# coding: utf-8

from . import constants as const
from .device import NukiDevice


class NukiOpener(NukiDevice):
    @property
    def is_rto_activated(self):
        return self.state == const.STATE_OPENER_RTO_ACTIVE

    @property
    def ring_action_timestamp(self):
        return self._json.get("ringactionTimestamp")

    @property
    def ring_action_state(self):
        return self._json.get("ringactionState")

    def activate_rto(self, block=False):
        return self._bridge.lock_action(
            nuki_id=self.nuki_id,
            action=const.ACTION_OPENER_ACTIVATE_RTO,
            device_type=self.device_type,
            block=block,
        )

    def deactivate_rto(self, block=False):
        return self._bridge.lock_action(
            nuki_id=self.nuki_id,
            action=const.ACTION_OPENER_DEACTIVATE_RTO,
            device_type=self.device_type,
            block=block,
        )

    def electric_strike_actuation(self, block=False):
        return self._bridge.lock_action(
            nuki_id=self.nuki_id,
            action=const.ACTION_OPENER_ELECTRIC_STRIKE_ACTUATION,
            device_type=self.device_type,
            block=block,
        )

    def activate_continuous_mode(self, block=False):
        return self._bridge.lock_action(
            nuki_id=self.nuki_id,
            action=const.ACTION_OPENER_ACTIVATE_CONTINUOUS,
            device_type=self.device_type,
            block=block,
        )

    def deactivate_continuous_mode(self, block=False):
        return self._bridge.lock_action(
            nuki_id=self.nuki_id,
            action=const.ACTION_OPENER_DEACTIVATE_CONTINUOUS,
            device_type=self.device_type,
            block=block,
        )



# coding: utf-8

# https://developer.nuki.io/page/nuki-bridge-http-api-1-10/4#heading--lock-actions
ACTION_LOCK_UNLOCK = 1
ACTION_LOCK_LOCK = 2
ACTION_LOCK_UNLATCH = 3
ACTION_LOCK_LOCK_N_GO = 4
ACTION_LOCK_LOCK_N_GO_WITH_UNLATCH = 5

ACTION_OPENER_ACTIVATE_RTO = 1
ACTION_OPENER_DEACTIVATE_RTO = 2
ACTION_OPENER_ELECTRIC_STRIKE_ACTUATION = 3
ACTION_OPENER_ACTIVATE_CONTINUOUS = 4
ACTION_OPENER_DEACTIVATE_CONTINUOUS = 5

# https://developer.nuki.io/page/nuki-bridge-http-api-1-10/4#heading--info
BRIDGE_TYPE_HW = 1
BRIDGE_TYPE_SW = 2

# https://developer.nuki.io/page/nuki-bridge-http-api-1-10/4#heading--device-types
DEVICE_TYPE_LOCK = 0
DEVICE_TYPE_OPENER = 2
DEVICE_TYPE_SMARTDOOR = 3
DEVICE_TYPE_SMARTLOCK3 = 4

# https://developer.nuki.io/page/nuki-bridge-http-api-1-10/4#heading--modes
MODE_LOCK_DOOR = 2
MODE_OPENER_DOOR = 2
MODE_OPENER_CONTINUOUS = 3

# https://developer.nuki.io/page/nuki-bridge-http-api-1-11/4/#heading--lock-states
STATE_LOCK_UNCALIBRATED = 0
STATE_LOCK_LOCKED = 1
STATE_LOCK_UNLOCKING = 2
STATE_LOCK_UNLOCKED = 3
STATE_LOCK_LOCKING = 4
STATE_LOCK_UNLATCHED = 5
STATE_LOCK_UNLOCKED_LOCK_N_GO = 6
STATE_LOCK_UNLATCHING = 7
STATE_LOCK_MOTOR_BLOCKED = 254
STATE_LOCK_UNDEFINED = 255

STATE_OPENER_UNTRAINED = 0
STATE_OPENER_ONLINE = 1
STATE_OPENER_RTO_ACTIVE = 3
STATE_OPENER_OPEN = 5
STATE_OPENER_OPENING = 7
STATE_OPENER_BOOT_RUN = 253
STATE_OPENER_UNDEFINED = 255

# https://developer.nuki.io/page/nuki-bridge-http-api-1-11/4/#heading--doorsensor-states
STATE_DOORSENSOR_DEACTIVATED = 1
STATE_DOORSENSOR_CLOSED = 2
STATE_DOORSENSOR_OPENED = 3
STATE_DOORSENSOR_UNKNOWN = 4
STATE_DOORSENSOR_CALIBRATING = 5



# coding: utf-8

from . import constants as const
from .device import NukiDevice


class NukiLock(NukiDevice):
    @property
    def is_locked(self):
        return self.state == const.STATE_LOCK_LOCKED

    @property
    def is_door_sensor_activated(self):
        # Nuki v1 locks don't have a door sensor, therefore the
        # door_sensor_state will is unset for them.
        if (
            not self.door_sensor_state
            or self.door_sensor_state == const.STATE_DOORSENSOR_UNKNOWN
        ):
            return
        return self.door_sensor_state != const.STATE_DOORSENSOR_DEACTIVATED

    @property
    def door_sensor_state(self):
        return self._json.get("doorsensorState")

    @property
    def door_sensor_state_name(self):
        return self._json.get("doorsensorStateName")

    @property
    def battery_charge(self):
        return self._json.get("batteryChargeState")

    @property
    def battery_critical_keypad(self):
        return self._json.get("keypadBatteryCritical")

    def lock(self, block=False):
        return self._bridge.lock(
            nuki_id=self.nuki_id, device_type=self.device_type, block=block
        )

    def unlock(self, block=False):
        return self._bridge.unlock(
            nuki_id=self.nuki_id, device_type=self.device_type, block=block
        )

    def lock_n_go(self, unlatch=False, block=False):
        return self._bridge.lock_n_go(
            nuki_id=self.nuki_id,
            unlatch=unlatch,
            device_type=self.device_type,
            block=block,
        )

    def unlatch(self, block=False):
        return self._bridge.unlatch(
            nuki_id=self.nuki_id, device_type=self.device_type, block=block
        )



# coding: utf-8

import hashlib
import logging
from datetime import datetime
from random import randint

import nacl.secret
import nacl.utils

logger = logging.getLogger("pynuki")


def sha256sum(text):
    return hashlib.sha256(text.encode("utf-8")).digest()


def encrypt_token(digest):
    nonce = nacl.utils.random(24)
    rnr = randint(0, 65535)
    ts = datetime.utcnow().strftime(f"%Y-%m-%dT%H:%M:%SZ,{rnr}")
    box = nacl.secret.SecretBox(digest)
    ctoken = box.encrypt(ts.encode("utf-8"), nonce)
    return {"ctoken": str(ctoken.ciphertext.hex()), "nonce": str(nonce.hex())}



#!/usr/bin/python
# coding: utf-8

"""
Based on the Bridge API version 1.13.2

Documentation:
https://developer.nuki.io/t/bridge-http-api/26
https://developer.nuki.io/uploads/short-url/a8eIacr0ku9zogOyuIuSEyw1PcA.pdf
"""

import logging

import requests

from . import constants as const
from .device import NukiDevice
from .lock import NukiLock
from .opener import NukiOpener
from .utils import encrypt_token, logger, sha256sum

# Default values
REQUESTS_TIMEOUT = 5


class InvalidCredentialsException(Exception):
    pass


class NukiBridge(object):
    def __init__(
        self,
        hostname,
        token=None,
        port=8080,
        secure=True,
        timeout=REQUESTS_TIMEOUT,
    ):
        self.hostname = hostname
        self.port = port
        self.__api_url = f"http://{hostname}:{port}"
        self.secure = secure
        self.requests_timeout = timeout
        self._json = None
        self.token = token

    def __repr__(self):
        return f"<NukiBridge: {self.hostname}:{self.port} (token={self.token})>"

    @staticmethod
    def discover():
        res = requests.get("https://api.nuki.io/discover/bridges")
        data = res.json()
        logger.debug(f"Discovery returned {data}")
        error_code = data.get("errorCode", -9999)
        if error_code != 0:
            logger.error(f"Discovery failed with error code {error_code}")
        bridges = data.get("bridges")
        if not bridges:
            logger.warning("No bridge discovered.")
        else:
            return [
                NukiBridge(hostname=x.get("ip"), port=x.get("port"))
                for x in bridges
            ]

    @property
    def token(self):
        return self.__token

    @property
    def tokendigest(self):
        return self.__tokendigest

    @token.setter
    def token(self, token):
        self.__token = token
        self.__tokendigest = sha256sum(token) if token else None
        # Try to log in if token has been set
        if self.token:
            try:
                self.info()
                logger.info("Login succeeded.")
            except requests.exceptions.HTTPError as err:
                if err.response.status_code == 401:
                    logger.error("Could not login with provided credentials")
                    raise InvalidCredentialsException(
                        "Login error. Provided token is invalid."
                    )

    @property
    def is_hardware_bridge(self):
        info = self.info()
        return info.get("bridgeType") == const.BRIDGE_TYPE_HW

    def __rq(self, endpoint, params=None):
        url = f"{self.__api_url}/{endpoint}"
        if self.secure:
            get_params = encrypt_token(self.tokendigest)
        else:
            get_params = {"token": self.token}
        if params:
            get_params.update(params)
        # Convert list to string to prevent request from encoding the url params
        # https://stackoverflow.com/a/23497912
        get_params_str = "&".join(f"{k}={v}" for k, v in get_params.items())
        result = requests.get(
            url, params=get_params_str, timeout=self.requests_timeout
        )
        result.raise_for_status()
        data = result.json()
        if "success" in data:
            if not data.get("success"):
                logger.warning(f"Call failed: {result}")
        return data

    def auth(self):
        url = f"{self.__api_url}/auth"
        result = requests.get(url, timeout=self.requests_timeout)
        result.raise_for_status()
        data = result.json()
        if not data.get("success", False):
            logging.warning(
                "Failed to authenticate against bridge. "
                "Have you pressed the button?"
            )
        return data

    def config_auth(self, enable):
        return self.__rq("configAuth", {"enable": 1 if enable else 0})

    def list(self, device_type=None):
        data = self.__rq("list")
        if device_type is not None:
            return [x for x in data if x.get("deviceType") == device_type]
        return data

    def lock_state(self, nuki_id, device_type=const.DEVICE_TYPE_LOCK):
        return self.__rq(
            "lockState", {"nukiId": nuki_id, "deviceType": device_type}
        )

    def lock_action(
        self, nuki_id, action, device_type=const.DEVICE_TYPE_LOCK, block=False
    ):
        params = {
            "nukiId": nuki_id,
            "deviceType": device_type,
            "action": action,
            "noWait": 0 if block else 1,
        }
        return self.__rq("lockAction", params)

    def unpair(self, nuki_id, device_type=const.DEVICE_TYPE_LOCK):
        return self.__rq(
            "unpair", {"nukiId": nuki_id, "deviceType": device_type}
        )

    def info(self):
        # Return cached value
        if self._json:
            return self._json
        data = self.__rq("info")
        self._json = data
        return data

    # Callback endpoints

    def callback_add(self, callback_url):
        return self.__rq("callback/add", {"url": callback_url})

    def callback_list(self):
        return self.__rq("callback/list")

    def callback_remove(self, callback_id):
        return self.__rq("callback/remove", {"id": callback_id})

    # Maintainance endpoints

    def log(self, offset=0, count=100):
        return self.__rq("log", {"offset": offset, "count": count})

    def clear_log(self):
        return self.__rq("clearlog")

    def firmware_update(self):
        return self.__rq("fwupdate")

    def reboot(self):
        return self.__rq("reboot")

    def factory_reset(self):
        return self.__rq("factoryReset")

    # Shorthand methods

    def update(self):
        # Invalidate cache
        self._json = None
        return self.info()

    def _get_devices(self, device_type=None):
        devices = []
        for dev in self.list(device_type=device_type):
            # lock_data holds the name and nuki id of the lock
            # eg: {'name': 'Home', 'nukiId': 241563832}
            device_data = {
                k: v for k, v in dev.items() if k not in ["lastKnownState"]
            }
            # state_data holds the last known state of the lock
            # eg: {'batteryCritical': False, 'state': 1, 'stateName': 'locked'}
            try:
                state_data = {
                    k: v
                    for k, v in dev["lastKnownState"].items()
                    if k not in ["timestamp"]
                }
            except KeyError:
                state_data = {}

            # Merge lock_data and state_data
            data = {**device_data, **state_data}

            dev_type = device_data.get("deviceType")
            if dev_type == const.DEVICE_TYPE_LOCK:
                dev = NukiLock(self, data)
            elif dev_type == const.DEVICE_TYPE_SMARTDOOR:
                dev = NukiLock(self, data)
            elif dev_type == const.DEVICE_TYPE_SMARTLOCK3:
                dev = NukiLock(self, data)
            elif dev_type == const.DEVICE_TYPE_OPENER:
                dev = NukiOpener(self, data)
            else:
                logger.warning(f"Unknown device type: {dev_type}")
                dev = NukiDevice(self, data)

            devices.append(dev)
        return devices

    @property
    def devices(self):
        return self._get_devices()

    @property
    def locks(self):
        locks = []
        for device in self._get_devices():
            if isinstance(device, NukiLock):
                locks.append(device)
        return locks

    @property
    def openers(self):
        return self._get_devices(device_type=const.DEVICE_TYPE_OPENER)

    def lock(self, nuki_id, device_type=const.DEVICE_TYPE_LOCK, block=False):
        return self.lock_action(
            nuki_id,
            action=const.ACTION_LOCK_LOCK,
            device_type=device_type,
            block=block,
        )

    def unlock(self, nuki_id, device_type=const.DEVICE_TYPE_LOCK, block=False):
        return self.lock_action(
            nuki_id,
            action=const.ACTION_LOCK_UNLOCK,
            device_type=device_type,
            block=block,
        )

    def lock_n_go(
        self,
        nuki_id,
        device_type=const.DEVICE_TYPE_LOCK,
        unlatch=False,
        block=False,
    ):
        action = const.ACTION_LOCK_LOCK_N_GO
        if unlatch:
            action = const.ACTION_LOCK_LOCK_N_GO_WITH_UNLATCH
        return self.lock_action(
            nuki_id,
            action=action,
            device_type=device_type,
            block=block,
        )

    def unlatch(self, nuki_id, device_type=const.DEVICE_TYPE_LOCK, block=False):
        return self.lock_action(
            nuki_id,
            action=const.ACTION_LOCK_UNLATCH,
            device_type=device_type,
            block=block,
        )

    def simple_lock(self, nuki_id, device_type=const.DEVICE_TYPE_LOCK):
        return self.__rq("lock", {"nukiId": nuki_id, "deviceType": device_type})

    def simple_unlock(self, nuki_id, device_type=const.DEVICE_TYPE_LOCK):
        return self.__rq(
            "unlock", {"nukiId": nuki_id, "deviceType": device_type}
        )



# coding: utf-8

from . import constants as const
from .exceptions import NukiUpdateException
from .utils import logger


class NukiDevice(object):
    def __init__(self, bridge, json):
        self._bridge = bridge
        self._json = json

    @property
    def name(self):
        return self._json.get("name")

    @property
    def nuki_id(self):
        return self._json.get("nukiId")

    @property
    def battery_critical(self):
        return self._json.get("batteryCritical")

    @property
    def firmware_version(self):
        return self._json.get("firmwareVersion")

    @property
    def state(self):
        return self._json.get("state")

    @property
    def state_name(self):
        return self._json.get("stateName")

    @property
    def device_type(self):
        return self._json.get("deviceType")

    @property
    def device_type_str(self):
        dev = self.device_type
        if dev == const.DEVICE_TYPE_LOCK:
            return "lock"
        elif dev == const.DEVICE_TYPE_OPENER:
            return "opener"
        logger.error(f"Unknown device type: {dev}")
        return "UNKNOWN"

    @property
    def device_model_str(self):
        dev = self.device_type
        if dev == const.DEVICE_TYPE_LOCK:
            return "Nuki Smart Lock 1.0/2.0"
        elif dev == const.DEVICE_TYPE_OPENER:
            return "Nuki Opener"
        elif dev == const.DEVICE_TYPE_SMARTDOOR:
            return "Nuki Smart Door"
        elif dev == const.DEVICE_TYPE_SMARTLOCK3:
            return "Nuki Smart Lock 3.0 (Pro)"
        logger.error(f"Unknown device type: {dev}")
        return "UNKNOWN"

    @property
    def mode(self):
        return self._json.get("mode")

    @property
    def mode_str(self):
        dev = self.device_type
        mode = self.mode
        if dev == const.DEVICE_TYPE_LOCK:
            if mode == const.MODE_LOCK_DOOR:
                return "door mode"
        elif dev == const.DEVICE_TYPE_OPENER:
            if mode == const.MODE_OPENER_DOOR:
                return "door mode"
            elif mode == const.MODE_OPENER_CONTINUOUS:
                return "continuous"
        logger.error(f'Unknown mode "{mode}" for device type {dev}')
        return "UNKNOWN"

    def update(self, aggressive=False):
        """
        Update the state of the Nuki device
        :param aggressive: Whether to aggressively poll the Bridge. If set to
        True, this will actively query the Lock, thus using more battery.
        :type aggressive: bool
        """
        if aggressive:
            data = self._bridge.lock_state(self.nuki_id)
            logger.debug(f"Received data: {data}")
            if not data.get("success", False):
                raise NukiUpdateException(
                    f"Failed to update data for Nuki device {self.nuki_id}"
                )
            self._json.update({k: v for k, v in data.items() if k != "success"})
        else:
            data = [
                dev
                for dev in self._bridge._get_devices(self.device_type)
                if dev.nuki_id == self.nuki_id
            ]
            assert data, (
                "Failed to update data for lock. "
                f"Nuki ID {self.nuki_id} volatized."
            )
            self._json.update(data[0]._json)

    def update_from_callback(self, json):
        """
        Update the state of the Nuki device from a callback
        :param json: Callback JSON body
        """
        assert (
            json.get("nukiId") == self.nuki_id
        ), "Failed to update data from callback. Wrong Nuki ID."
        self._json.update(json)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self._json}>"
